/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.definition.activity;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.Execution;
import org.jbpm.pvm.activity.ActivityExecution;
import org.jbpm.pvm.activity.ExternalActivity;
import org.jbpm.pvm.internal.env.Authentication;
import org.jbpm.pvm.internal.model.NodeImpl;
import org.jbpm.pvm.internal.model.TransitionImpl;
import org.jbpm.pvm.internal.model.VariableDefinitionImpl;
import org.jbpm.pvm.model.Condition;
import org.jbpm.pvm.model.Transition;
import org.ow2.bonita.definition.ClassInfo;
import org.ow2.bonita.definition.JavaHook;
import org.ow2.bonita.definition.MultiInstantiator;
import org.ow2.bonita.definition.MultiInstantiatorDescriptor;
import org.ow2.bonita.definition.Performer;
import org.ow2.bonita.facade.exception.BonitaWrapperException;
import org.ow2.bonita.facade.exception.MultiInstantiatorInvocationException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.JoinRunTime;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.util.AccessorUtil;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.novabpm.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras, Pascal Verdage
 */
/**
 * Activity life cycle:
 * 1- when an instance is created, activity state is {@link ActivityState#INITIAL}
 * 2- when the execution arrives on the node, the activity becomes {@link ActivityState#READY}
 *  the activity state is recorded (before start)
 * 3- if the activity is Manual, a task is created.
 * - when the start task is started, the activity state is recorded (after start).
 * 4- the activity becomes {@link ActivityState#EXECUTING}. The business logic is executed.
 * - the activity state is recorded (before stopped)
 * 5- if the activity is Manual, wait for the task to finish. (TODO: change the activity state ?)
 * - when the task is finished, the activity state is recorded (after stopped).
 * 6- the activity becomes {@link ActivityState#FINISHED}.
 */
public abstract class AbstractActivity implements ExternalActivity {

  public static enum ActivityType {
    route, block, no, tool, subFlow, task
  }

  public static enum SplitType {
    AND, XOR
  }

  public static enum JoinType {
    AND, XOR
  }

  /** LOG */
  private static final Logger LOG = Logger.getLogger(AbstractActivity.class.getName());
  protected long dbid;

  protected String activityId;
  protected SplitType splitType;
  protected JoinType joinType;

  protected ActivityType activityType;
  protected List<VariableDefinitionImpl> variableDefinitions;
  protected Performer performer;

  protected Set<String> deadlines;

  protected List<JavaHook> javaHooks;
  // multi instantiation
  protected ClassInfo multiInstantiationClass;
  protected String multiInstantiationVariable;
  
  public static final String BODY_FINISHED = "bodyFinished";
  public static final String ACT_INSTANCE_FINISHED = "instFinished";


  protected AbstractActivity() {
  }

  protected AbstractActivity(String activityName, JoinType joinType,
      SplitType splitType,
      ActivityType activityType,
      Performer performer) {

    Misc.checkArgsNotNull(activityType);
    this.activityId = activityName;
    this.joinType = joinType;
    this.splitType = splitType;
    this.activityType = activityType;
    this.performer = performer;

    if (performer != null) {
      if (!ActivityType.task.equals(activityType)) {
        throw new BonitaRuntimeException("Activity is automatic and a performer is specified: incorrect.");
      }
    } else {
      if (ActivityType.task.equals(activityType)) {
        throw new BonitaRuntimeException("startMode or finishMode is Manual and no performer is specified: incorrect.");
      }
    }
  }
  
  public void execute(ActivityExecution execution) {
    XpdlExecution xpdlExecution = (XpdlExecution) execution;
    // If instance is ended, don't execute next node.
    if (xpdlExecution.getXpdlInstance().getInstanceState().equals(InstanceState.FINISHED)) {
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("XPDLInstance ended : " + xpdlExecution.getXpdlInstance());
      }
      xpdlExecution.end();
      XpdlExecution parent = xpdlExecution.getParent();
      if (parent != null) {
        parent.removeExecution(xpdlExecution);
      }
      return;
    }
    // Execute node.
    boolean joinOK = true;
    joinOK = executeJoin(execution);
    if (joinOK) {
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("Join for activity " + this + " is OK.");
      }
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("Crating a new iteration on activity : " + this);
      }
      createNewIteration(xpdlExecution);

      String nodeName = execution.getNode().getName();

      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("Executing node: " + nodeName + ", class = " + this.getClass().getSimpleName());
      }

      if (multiInstantiationClass != null) {
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("MultiInstantiation not null on activity " + this);
        }
        MultiInstantiatorDescriptor actInstDescr = null;

        ProcessInstanceUUID instanceUUID = xpdlExecution.getXpdlInstance().getUUID();
        MultiInstantiator actInstantiator = 
          ClassDataLoader.getActivityInstantiatorInstance(
              xpdlExecution.getXpdlInstance().getPackageDefinitionUUID(), multiInstantiationClass);
        try {
          actInstDescr = actInstantiator.execute(
                AccessorUtil.getQueryAPIAccessor(), 
                instanceUUID, activityId, xpdlExecution.getIterationId());
          if (actInstDescr == null) {
            throw new BonitaRuntimeException("MultiInstantiator execution returned null in activity " + activityId);
          }
        } catch (Exception e) {
          throw new BonitaWrapperException(
              new MultiInstantiatorInvocationException(multiInstantiationClass.toString(), e)
          );
        } 

        xpdlExecution.setWaitingForActivityInstanceNb(actInstDescr.getJoinNumber());
        int childId = 0;
        for (Object value : actInstDescr.getVariableValues()) {
          XpdlExecution childExec = createChildExecution(xpdlExecution, childId);
          childExec = (XpdlExecution) childExec.createScope(xpdlExecution.getNode());
          childExec.setVariable(multiInstantiationVariable, value);
          childExec.setActivityInstanceId(Integer.toString(childId));
          executeActivityInstance(childExec);
          childId++;
        }
      } else {
        xpdlExecution = (XpdlExecution) xpdlExecution.createScope(xpdlExecution.getNode());
        executeActivityInstance(xpdlExecution);
      }
    } else {
      xpdlExecution.end();
      XpdlExecution parent = xpdlExecution.getParent();
      if (parent != null) {
        parent.removeExecution(xpdlExecution);
      }
    }
  }

  protected void executeActivityInstance(XpdlExecution xpdlExecution) {

    final Recorder recorder = EnvTool.getRecorder();
    recorder.recordEnterActivity(this, xpdlExecution);

    boolean canContinue = executeBody(xpdlExecution);
    if (canContinue) {
      end(xpdlExecution);
    } else {
      xpdlExecution.waitForSignal();
    }    
  }
  
  protected void end(XpdlExecution xpdlExecution) {
    EnvTool.getRecorder().recordBodyEnded(this, xpdlExecution);
    if (!XpdlExecution.MAIN_INSTANCE_NAME.equals(xpdlExecution.getActivityInstanceId())) {
      // TODO: check variable values for transition conditions
      XpdlExecution activityInstanceExecution = (XpdlExecution) xpdlExecution.destroyScope(xpdlExecution.getNode());
      activityInstanceExecution.end();
      XpdlExecution parent = activityInstanceExecution.getParent();
      parent.removeExecution(activityInstanceExecution);
      signal(parent, ACT_INSTANCE_FINISHED, null);
    } else {
      executeSplit(xpdlExecution, true);
    }
  }

  public void signal(ActivityExecution execution, String signal, Map<String, Object> parameters) {
    XpdlExecution xpdlExecution = (XpdlExecution) execution;
    if (BODY_FINISHED.equals(signal)) {
      end(xpdlExecution);
    } else if (ACT_INSTANCE_FINISHED.equals(signal)) {
      xpdlExecution.setWaitingForActivityInstanceNb(xpdlExecution.getWaitingForActivityInstanceNb() - 1);
      if (xpdlExecution.getWaitingForActivityInstanceNb() == 0) {
        // TODO: cancel remaining activity instances
        executeSplit(xpdlExecution, false);
      }
    } else if (signal != null && deadlines != null) {
      for (String deadline : deadlines) {
        if (deadline.equals(signal)) {
          Authentication.setUserId("SYSTEM");
          // By default, a deadline does not propagate execution
          xpdlExecution.waitForSignal();
          String activityId = xpdlExecution.getNode().getName();
          JavaHook javaHook = new JavaHook(signal, JavaHook.Type.onDeadline, true);
          HookExecutor.executeHook(xpdlExecution, activityId, javaHook);
          return;
        }
      }
    }
  }

  protected void executeSplit(Execution execution, boolean removeScope) {
    XpdlExecution xpdlExecution = (XpdlExecution) execution;
    NodeImpl currentNode = xpdlExecution.getNode();
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("node = " + currentNode.getName() + " - splitType = "
          + splitType + " - execution = " + execution.getName());
    }
    List<Transition> transitions = currentNode.getOutgoingTransitions();
    if (transitions == null) {
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("node = " + currentNode.getName() + " - splitType = "
            + splitType + " - execution = " + execution.getName() + " no transition available. Ending execution");
      }
      xpdlExecution.end();
      XpdlExecution parent = xpdlExecution.getParent();
      if (parent != null) {
        parent.removeExecution(xpdlExecution);
      }
    } else {
      List<Transition> transitionsToTake = new ArrayList<Transition>();
      for (Transition t : transitions) {
        if (evaluateTransition((TransitionImpl) t, xpdlExecution)) {
          transitionsToTake.add(t);
        }
      }
      // remove not propagated variables
      if (removeScope) {
        xpdlExecution = (XpdlExecution) xpdlExecution.destroyScope(currentNode);
      }
      if (transitionsToTake.size() == 0) {
        xpdlExecution.end();
        XpdlExecution parent = xpdlExecution.getParent();
        if (parent != null) {
          parent.removeExecution(xpdlExecution);
        }        
      } else if (transitionsToTake.size() == 1  || splitType.equals(SplitType.XOR)) {
        // We are in a Split/AND and only one transition is true,
        // or we are in a Split/XOR, so we take the first one that is true.
        Transition t = transitionsToTake.get(0);
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("Taking transition " + t);
        }
        xpdlExecution.take(t);
      } else {
        // We are in a Split/AND and more than one transition is true.
        int childId = 0;
        for (Transition t : transitionsToTake) {
          XpdlExecution childExecution = createChildExecution(xpdlExecution, childId++);
          if (LOG.isLoggable(Level.FINE)) {
            LOG.fine("Execution " + childExecution.getName() + "is taking transition " + t);
          }
          childExecution.take(t);
        }
      }
    }
  }

  private boolean evaluateTransition(TransitionImpl t, XpdlExecution xpdlExecution) {
    // testing the guard condition
    Condition condition = t.getCondition();
    ConditionDefinition condDef = (ConditionDefinition) condition;
    boolean conditionOK = true;
    if (condDef != null) {
      conditionOK = condDef.evaluate(xpdlExecution);
      if (!conditionOK) {
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("Unable to take transition: " + t.getName());
        }
        // TODO : cancel nodes on this branch : may be investigate for cleaning of executions ?
      }
    }
    return conditionOK;
  }

  protected XpdlExecution createChildExecution(XpdlExecution parentExecution, int childId) {
    String executionName = parentExecution.getName() + "/" + childId;
    XpdlExecution childExecution = (XpdlExecution) parentExecution.createExecution(executionName);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Creating child execution with name " + childExecution.getName());
    }
    return childExecution;
  }

  protected boolean executeJoin(Execution execution) {
    XpdlExecution currentXpdlExecution = (XpdlExecution) execution;
    JoinRunTime joinRunTime = getJoinRuntime(currentXpdlExecution);

    return joinRunTime.transitionArrived(
        currentXpdlExecution.getPreviousTransition().getSource().getName(),
        currentXpdlExecution.getIterationId());
  }

  private static JoinRunTime getJoinRunTime(XpdlInstance instance, NodeImpl node) {
    JoinType joinType = ((AbstractActivity) node.getBehaviour()).getJoinType();
    JoinRunTime joinRunTime = instance.getJoinRunTime(node.getName());
    if (joinRunTime == null) {
      int joinIncomingTransitionNbRequired = 1;
      if (joinType.equals(JoinType.AND)) {
        joinIncomingTransitionNbRequired = node.getIncomingTransitions().size();
      }
      Map<String, JoinRunTime.WaitingTransition> ins = new HashMap<String, JoinRunTime.WaitingTransition>();
      for (Transition t : node.getIncomingTransitions()) {
        String name = t.getSource().getName();
        ins.put(name, new JoinRunTime.WaitingTransition(name, null));
      }
      joinRunTime = new JoinRunTime(node.getName(), ins, joinIncomingTransitionNbRequired);
      instance.pushJoinRuntime(joinRunTime);
    }
    return joinRunTime;
  }

  private static JoinRunTime getJoinRuntime(XpdlExecution xpdlExecution) {
    return getJoinRunTime(xpdlExecution.getXpdlInstance(), xpdlExecution.getNode());
  }



  public static class IterationDescriptor {
    private String joinNodeName;
    private Set<String> transitionSourceNodes;

    protected IterationDescriptor() { }
    public IterationDescriptor(String joinNodeName, Set<String> transitionSourceNodes) {
      this.joinNodeName = joinNodeName;
      this.transitionSourceNodes = transitionSourceNodes;
    }
    public String getJoinNodeName() {
      return joinNodeName;
    }
    public Set<String> getTransitionSourceNodes() {
      return transitionSourceNodes;
    }
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
      return "for join " + joinNodeName + ", reinitialize transitions from " + transitionSourceNodes;
    }
  }

  private List<IterationDescriptor>  iterationDescriptors;

  public List<IterationDescriptor> getIterationDescriptors() {
    return iterationDescriptors;
  }

  public void setIterationDescriptors(List<IterationDescriptor> iteration) {
    this.iterationDescriptors = iteration;
  }

  private void createNewIteration(XpdlExecution execution) {
    if (iterationDescriptors != null && !iterationDescriptors.isEmpty()) {
      String iterationUUID = Misc.getUniqueId("it");
      execution.setIterationId(iterationUUID);
      XpdlInstance instance = execution.getXpdlInstance();
      for (IterationDescriptor it : iterationDescriptors) {
        JoinRunTime jr = getJoinRunTime(instance, execution.getProcessDefinition().getNode(it.getJoinNodeName()));
        if (jr.update(iterationUUID, it.getTransitionSourceNodes())) {
          throw new BonitaRuntimeException("Error in iteration: join " + it.getJoinNodeName() + " has not been reinitialized");
        }
      }
    }
  }

  private boolean executeBody(XpdlExecution execution) {
    if (bodyStartAutomatically()) {
      EnvTool.getRecorder().recordBodyStarted(this, execution);
    }
    return executeBusinessLogic(execution);
  }

  public abstract ActivityBody getBody(XpdlExecution xpdlExecution, ActivityInstance<ActivityBody> activityInstance);

  /**
   * Return true if the execution can continue
   */
  protected abstract boolean executeBusinessLogic(Execution execution);
  protected abstract boolean bodyStartAutomatically();
  public List<VariableDefinitionImpl> getVariableDefinitions() {
    return variableDefinitions;
  }

  public void setVariableDefinitions(
      List<VariableDefinitionImpl> variableDefinitions) {
    this.variableDefinitions = variableDefinitions;
  }


  public ActivityType getActivityType() {
    return activityType;
  }


  public Performer getPerformer() {
    return performer;
  }

  public JoinType getJoinType() {
    return joinType;
  }


  public void setJoinType(JoinType joinType) {
    this.joinType = joinType;
  }

  public SplitType getSplitType() {
    return splitType;
  }

  public Set<String> getDeadlineHooks() {
    return deadlines;
  }

  public void setDeadlineHooks(Set<String> deadlineHooks) {
    this.deadlines = deadlineHooks;
  }

  public void setJavaHooks(List<JavaHook> javaHooks) {
    this.javaHooks = javaHooks;
  }


  public List<JavaHook> getJavaHooks() {
    return javaHooks;
  }


  public String getActivityId() {
    return activityId;
  }

  
  public void setMultiInstantiaton(String variableId, ClassInfo instantiator) {
    this.multiInstantiationVariable = variableId;
    this.multiInstantiationClass = instantiator;
  }
}
