package org.ow2.bonita.env;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.env.EnvironmentFactory;
import org.ow2.bonita.facade.Context;
import org.ow2.bonita.persistence.db.DbJournal;
import org.ow2.bonita.persistence.db.DbRepository;
import org.ow2.bonita.persistence.log.LoggerArchiver;
import org.ow2.bonita.persistence.log.LoggerRecorder;
import org.ow2.bonita.persistence.xml.XMLHistory;
import org.ow2.bonita.runtime.Repository;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.services.handlers.FinishedInstanceHandler;
import org.ow2.bonita.services.handlers.UndeployedPackageHandler;
import org.ow2.bonita.services.handlers.impl.ArchiveFinishedInstanceHandler;
import org.ow2.bonita.services.handlers.impl.ArchiveUndeployedPackageHandler;
import org.ow2.bonita.services.handlers.impl.DeleteFinishedInstanceHandler;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.novabpm.util.Misc;
import org.ow2.novabpm.util.Misc.NullCheckResult;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class EnvGenerator {

  private static final Logger LOG = Logger.getLogger(EnvGenerator.class.getName());

  public static final String RETRY = "retry";
  
  

  public static final String API_CONFIG_KEY = "api-config";

  public static final String HISTORY_DEFAULT_KEY = "history";
  public static final String JOURNAL_DEFAULT_KEY = "journal";
  
  private Map<String, String> applicationVariables;

  private Map<String, String> blockVariables;

  private String repositoryXml = null;

  private String finishedInstanceHandlerXml = null;

  private String undeployedPackageHandlerXml = null;

  private String recorderXml = null;

  private String journalXml = null;

  private String archiverXml = null;

  private String historyXml = null;

  protected List<String> mappingFiles = new ArrayList<String>();

  private String apiTypeXml = null;

  protected static String defaultIndent = "  ";

  @SuppressWarnings("unchecked")
  public EnvGenerator() {
    applicationVariables = new HashMap<String, String>();
    blockVariables = new HashMap<String, String>();

    this.setAPIType(Context.AutoDetect);

    // Hibernate configuration for bonita core engine.
    setRuntimeHibernateConf();
    


    // Command execution configuration
    String commandServiceConfigXml = "<command-service>\n";
    commandServiceConfigXml += defaultIndent + "<retry-interceptor />\n";
    commandServiceConfigXml += defaultIndent + "<environment-interceptor />\n";
    commandServiceConfigXml += defaultIndent + "<standard-transaction-interceptor />\n";
    commandServiceConfigXml += "</command-service>\n";
    this.setApplicationEntry("command-service", commandServiceConfigXml);
    this.setApplicationEntry("job-executor", "<job-executor threads='1' auto-start='true' />");
    this.setApplicationEntry("type-resolver", "<variable-types resource='bonita.type.resolver.xml' />");

    this.setRepositoryType(DbRepository.class, "<arg><string value='" 
        + EnvConstants.CORE_PERSISTENCESERVICE_NAME + "' /></arg>");

    this.setBlockEntry("transaction", "<transaction />");

    this.setJournalType(DbJournal.class, "<arg><string value='" 
        + EnvConstants.CORE_PERSISTENCESERVICE_NAME + "' /></arg>");

    this.setFinishedInstanceHandlerType(DeleteFinishedInstanceHandler.class, ArchiveFinishedInstanceHandler.class);

    this.setUndeployedPackageHandlerType(ArchiveUndeployedPackageHandler.class);

    this.setRecorderType(LoggerRecorder.class);
    this.setArchiverType(LoggerArchiver.class);
    this.setHistoryType(XMLHistory.class);
    this.setBlockEntry("timer_session", "<timer-session />");
    this.setBlockEntry("job-session", "<job-db-session session='" + EnvConstants.CORE_HB_SESSION_NAME + "' />");
  }

  private void setRuntimeHibernateConf() {
    String hibernateConfigXml = "";
    hibernateConfigXml += "<hibernate-configuration name='" + EnvConstants.CORE_HB_CONF_NAME + "' >\n";
    hibernateConfigXml += defaultIndent + "<properties resource='hibernate-core.properties' />\n";
    hibernateConfigXml += defaultIndent + "<mappings resource='bonita.mappings.hbm.xml' />\n";
    hibernateConfigXml += defaultIndent + "<cache-configuration resource='bonita.cache.xml' usage='read-write' />\n";
    hibernateConfigXml += "</hibernate-configuration>";
    this.setApplicationEntry(EnvConstants.CORE_HB_CONF_NAME, hibernateConfigXml);
    
    
    this.setApplicationEntry(EnvConstants.CORE_HB_SESSION_FACTORY_NAME, 
        "<hibernate-session-factory name='" 
        + EnvConstants.CORE_HB_SESSION_FACTORY_NAME
        + "' configuration='" + EnvConstants.CORE_HB_CONF_NAME + "' />");

    this.setBlockEntry(EnvConstants.CORE_HB_SESSION_NAME, 
        "<hibernate-session name='" + EnvConstants.CORE_HB_SESSION_NAME 
        + "' factory='" + EnvConstants.CORE_HB_SESSION_FACTORY_NAME + "' />");
    this.setBlockEntry(EnvConstants.CORE_PERSISTENCESERVICE_NAME, 
        "<" + EnvConstants.XPDL_PERSISTENCE_SERVICE + " name='" + EnvConstants.CORE_PERSISTENCESERVICE_NAME 
        + "' session='" + EnvConstants.CORE_HB_SESSION_NAME + "'/>");
  }
  
  private String getIndent(int depth) {
    String res = "";
    for (int i = 0; i <= depth; i++) {
      res += defaultIndent;
    }
    return res;
  }


  private String reIndent(int depth, String initialString) {
    StringReader stringReader = new StringReader(initialString);
    BufferedReader reader = new BufferedReader(stringReader);
    StringBuilder builder = new StringBuilder();
    String line;
    try {
      while ((line = reader.readLine()) != null) {
        if (builder.length() != 0) {
          builder.append(Misc.LINE_SEPARATOR);
        }
        builder.append(getIndent(depth)).append(line);
      }
    } catch (IOException e) {
      Misc.unreachableStatement();
    } finally {
      Misc.close(reader);
      Misc.close(stringReader);
    }

    return builder.toString();
  }
  /**
   * Generates a xml for the given environment configuration.
   *
   * @return
   */
  public String createEnvironmentXml() {
    final NullCheckResult nullCheckResult = Misc.findNull(repositoryXml,
        finishedInstanceHandlerXml,
        recorderXml,
        journalXml,
        archiverXml);

    if (nullCheckResult.hasNull()) {
      throw new BonitaRuntimeException("Error while creating environment xml: "
          + Misc.getStringFrom(nullCheckResult, "repository",
              "finishedInstanceHandlerType",
              "recorderType",
              "journalType", "archiverType") + " is/are null!");
    }

    int indentDepth = 0;


    // ENV XML GENERATION BEGINS HERE
    final StringBuilder result =
      new StringBuilder("<environment-definition>" + Misc.LINE_SEPARATOR + Misc.LINE_SEPARATOR);

    // DEFINITION OF THE ENVIRONMENT FACTORY
    indentDepth++;
    result.append(getIndent(indentDepth) + "<environment-factory>" + Misc.LINE_SEPARATOR);

    indentDepth++;
    for (String entry : applicationVariables.values()) {
      result.append(reIndent(indentDepth, entry)).append(Misc.LINE_SEPARATOR);
    }

    result
    .append(reIndent(indentDepth, apiTypeXml)).append(Misc.LINE_SEPARATOR)
    .append(reIndent(indentDepth, finishedInstanceHandlerXml)).append(Misc.LINE_SEPARATOR)
    .append(reIndent(indentDepth, undeployedPackageHandlerXml)).append(Misc.LINE_SEPARATOR);

    indentDepth--;
    result  
    .append(getIndent(indentDepth)).append("</environment-factory>" + Misc.LINE_SEPARATOR + Misc.LINE_SEPARATOR);
    // END OF ENVIRONMENT FACTORY
    // DEFINITION OF THE ENVIRONMENT

    result
    .append(getIndent(indentDepth)).append("<environment>" + Misc.LINE_SEPARATOR);
    indentDepth++;

    // journal
    result
    .append(reIndent(indentDepth, journalXml)).append(Misc.LINE_SEPARATOR);

    // history
    result
    .append(reIndent(indentDepth, historyXml)).append(Misc.LINE_SEPARATOR);

    // recorder
    result
    .append(getIndent(indentDepth)).append("<!-- DbJournal cannot be shared by several environments." + Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth)).append("It contains a session cache that needs to be recreated for each environment -->") 
    .append(Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth)).append("<chainer name='").append(Recorder.DEFAULT_KEY + "'>" + Misc.LINE_SEPARATOR)
    .append(reIndent(indentDepth + 1, recorderXml)).append(Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth + 1)).append("<ref object='" + JOURNAL_DEFAULT_KEY + "' />").append(Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth)).append("</chainer>" + Misc.LINE_SEPARATOR);

    // archiver
    result
    .append(getIndent(indentDepth)).append("<chainer name='").append(Archiver.DEFAULT_KEY + "'>" + Misc.LINE_SEPARATOR)
    .append(reIndent(indentDepth + 1 , archiverXml)).append(Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth + 1)).append("<ref object='" + HISTORY_DEFAULT_KEY + "' />").append(Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth)).append("</chainer>" + Misc.LINE_SEPARATOR);

    // query api
    result
    .append(getIndent(indentDepth)).append("<!-- Query Api has an object reference to the journal,").append(Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth)).append("so it cannot be shared by multiple environments -->").append(Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth)).append("<queryApi name='" + Querier.DEFAULT_KEY + "'>").append(Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth + 1)).append("<ref object='" + JOURNAL_DEFAULT_KEY + "' />").append(Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth + 1)).append("<ref object='" + HISTORY_DEFAULT_KEY + "' />").append(Misc.LINE_SEPARATOR)
    .append(getIndent(indentDepth)).append("</queryApi>").append(Misc.LINE_SEPARATOR);

    // repository
    result
    .append(getIndent(indentDepth)).append("<!-- DbRepository cannot be shared "
        + "by several environments.").append(Misc.LINE_SEPARATOR)
        .append(getIndent(indentDepth)).append("It contains a session cache that needs to be recreated for each environment -->" 
            + Misc.LINE_SEPARATOR)
            .append(reIndent(indentDepth, repositoryXml)).append(Misc.LINE_SEPARATOR);

    // other entries
    for (String entry : blockVariables.values()) {
      result.append(reIndent(indentDepth, entry)).append(Misc.LINE_SEPARATOR);
    }

    indentDepth--;
    result.append(getIndent(indentDepth)).append("</environment>" + Misc.LINE_SEPARATOR + Misc.LINE_SEPARATOR);
    indentDepth--;
    result.append(getIndent(indentDepth)).append("</environment-definition>" + Misc.LINE_SEPARATOR);

    return new String(result);
  }

  /**
   * Creates a new EnvironmentFactory from the list of entries.
   */
  public EnvironmentFactory createEnvironmentFactory() {
    final String envConfig = createEnvironmentXml();
    if (LOG.isLoggable(Level.CONFIG)) {
      LOG.config("The following environment has been generated by: "
          + this.getClass().getName() + Misc.LINE_SEPARATOR + envConfig);
    }
    EnvironmentFactory environmentFactory = XpdlEnvironmentParser
    .parseEnvironmentFactoryFromXmlString(envConfig);
    return environmentFactory;
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param classToUse
   *            classToUse at runtime
   * @return a string that declares an object to be added in the context.
   */
  @SuppressWarnings("unchecked")
  private static String getObjectDecl(final Class classToUse) {
    return "<object class='" + classToUse.getName() + "' />";
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param key
   *            objectName
   * @param classToUse
   *            classToUse at runtime
   * @return a string that declares an object to be added in the context.
   */
  @SuppressWarnings("unchecked")
  private static String getObjectDecl(final String key, final Class classToUse) {
    return "<object name='" + key + "' class='" + classToUse.getName() + "' />";
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param key
   *            objectName
   * @param classToUse
   *            classToUse at runtime
   * @param fieldKeys
   *            fields name
   * @param refKeys
   *            the keys of objects to reference in fields declaration
   * @return a string that declares an object to be added in the context.
   */
  @SuppressWarnings("unchecked")
  private static String getObjectDeclWithFields(final String key, final Class classToUse,
      String[] fieldKeys, String[] refKeys) {
    if (fieldKeys.length != refKeys.length) {
      throw new RuntimeException("Error : fieldKeys and refKeys don't have the same number of elements");
    }
    String result = "<object name='" + key + "' class='" + classToUse.getName() + "' >\n";
    for (int i = 0; i < fieldKeys.length; i++) {
      result += defaultIndent + "<field name='" + fieldKeys[i] + "'><ref object='" + refKeys[i] + "' /></field>\n";
    }
    result += "</object>";
    return result;
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param xml
   *            declaration of this entry
   * @return previous declaration for this name
   */
  public String setApplicationEntry(final String name, final String xml) {
    return applicationVariables.put(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @return previous declaration for this name
   */
  public String setApplicationEntry(final String name, final Class< ? > classToUse) {
    String xml = getObjectDecl(name, classToUse);
    return setApplicationEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @param fieldKeys
   *            list of fields to set
   * @param refKeys
   *            value to set to the fields
   * @return previous declaration for this name
   */
  public String setApplicationEntry(final String name, final Class< ? > classToUse,
      String[] fieldKeys, String[] refKeys) {
    String xml = getObjectDeclWithFields(name, classToUse, fieldKeys, refKeys);
    return setApplicationEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param xml
   *            declaration of this entry
   * @return previous declaration for this name
   */
  public String setBlockEntry(final String name, final String xml) {
    return blockVariables.put(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @return previous declaration for this name
   */
  public String setBlockEntry(final String name, final Class< ? > classToUse) {
    String xml = getObjectDecl(name, classToUse);
    return setBlockEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @param fieldKeys
   *            list of fields to set
   * @param refKeys
   *            value to set to the fields
   * @return previous declaration for this name
   */
  public String setBlockEntry(final String name, final Class< ? > classToUse,
      final String[] fieldKeys, final String[] refKeys) {
    String xml = getObjectDeclWithFields(name, classToUse, fieldKeys, refKeys);
    return setBlockEntry(name, xml);
  }

  /**
   * Prints the default environment on standard output.
   *
   * @param args
   * @throws IOException 
   */
  @SuppressWarnings("unchecked")
  public static void main(String[] args) throws IOException {
    if (args.length != 1 && args.length != 2) {
      System.exit(1);
    }
    EnvGenerator envGenerator = null;
    try {
      Class <EnvGenerator> envGeneratorClass =
        (Class<EnvGenerator>) EnvGenerator.class.getClassLoader().loadClass(args[0]);
      envGenerator = envGeneratorClass.newInstance();
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
    if (args.length == 1) {
      if (LOG.isLoggable(Level.INFO)) {
        LOG.config(envGenerator.createEnvironmentXml());
      }
    } else {
      Misc.write(envGenerator.createEnvironmentXml(), new File(args[1]));
    }
  }

  public void setFinishedInstanceHandlerType(final Class< ? extends FinishedInstanceHandler>... finishedInstanceHandlerTypes) {
    this.finishedInstanceHandlerXml = "<chainer name='" + FinishedInstanceHandler.DEFAULT_KEY + "'>" + Misc.LINE_SEPARATOR;
    for (Class< ? extends FinishedInstanceHandler> finishedInstanceHandlerType : finishedInstanceHandlerTypes) {
      this.finishedInstanceHandlerXml += defaultIndent + getObjectDecl(finishedInstanceHandlerType) + Misc.LINE_SEPARATOR;
    }
    this.finishedInstanceHandlerXml += "</chainer>" + Misc.LINE_SEPARATOR;
  }

  public void setUndeployedPackageHandlerType(final Class< ? extends UndeployedPackageHandler>... undeployedPackageHandlerTypes) {
    this.undeployedPackageHandlerXml = "<chainer name='" + UndeployedPackageHandler.DEFAULT_KEY + "'>" + Misc.LINE_SEPARATOR;
    for (Class< ? extends UndeployedPackageHandler> undeployedPackageHandlerType : undeployedPackageHandlerTypes) {
      this.undeployedPackageHandlerXml += defaultIndent + getObjectDecl(undeployedPackageHandlerType) + Misc.LINE_SEPARATOR;
    }
    this.undeployedPackageHandlerXml += "</chainer>" + Misc.LINE_SEPARATOR;
  }

  public void setRecorderType(final Class< ? extends Recorder> recorderType) {
    this.recorderXml = "<recorder class='" + recorderType.getName() + "' />";
  }

  public void setJournalType(final Class< ? extends Recorder> journalType, String... args) {
    checkExtendingQuerier(journalType);
    this.journalXml = "<journal name='" + JOURNAL_DEFAULT_KEY + "' class='" + journalType.getName() + "'> " + Misc.LINE_SEPARATOR;
    for (String arg : args) {
      this.journalXml += defaultIndent + arg + Misc.LINE_SEPARATOR;
    }
    this.journalXml += "</journal>";

  }

  private void setArchiverType(final Class< ? extends Archiver> archiverType) {
    this.archiverXml = "<archiver class='" + archiverType.getName() + "' />";
  }

  public void setHistoryType(final Class< ? extends Archiver> historyType, String... historyArgs) {
    checkExtendingQuerier(historyType);
    this.historyXml = "<history name='" + HISTORY_DEFAULT_KEY + "' class='" + historyType.getName() + "'>" + Misc.LINE_SEPARATOR;
    for (String arg : historyArgs) {
      this.historyXml += defaultIndent + arg + Misc.LINE_SEPARATOR;
    }
    this.historyXml += "</history>";
  }
  
  private void checkExtendingQuerier(Class < ? > clazz) {
    if (!(Querier.class.isAssignableFrom(clazz))) {
      throw new BonitaRuntimeException(
        "Invalid class " + clazz.getName() + " : it doesn't implement " + Querier.class + " interface.");
    }
  }

  public void setRepositoryType(final Class< ? extends Repository> repositoryType, String... args) {
    this.repositoryXml = "<repository class='" + repositoryType.getName() + "'>" + Misc.LINE_SEPARATOR;
    for (String arg : args) {
      this.repositoryXml += defaultIndent + arg + Misc.LINE_SEPARATOR;
    }
    this.repositoryXml += "</repository>";
  }

  public void setAPIType(final Context context) {
    this.apiTypeXml  = "<api " + " type='" + context.name() +  "' />";
  }

  public List<String> getMappingFiles() {
    return mappingFiles;
  }
}
