/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade;

import java.io.Serializable;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.security.Principal;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.naming.InitialContext;

import org.jbpm.pvm.PvmException;
import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.env.EnvironmentFactory;
import org.jbpm.pvm.internal.cmd.CommandService;
import org.jbpm.pvm.internal.env.Authentication;
import org.ow2.bonita.env.GlobalEnvironmentFactory;
import org.ow2.bonita.env.InvalidEnvironmentException;
import org.ow2.bonita.facade.exception.BonitaInternalException;
import org.ow2.bonita.facade.exception.BonitaWrapperException;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.novabpm.identity.auth.SecurityContext;
import org.ow2.novabpm.util.Command;
import org.ow2.novabpm.util.Misc;

public class APIInterceptor implements InvocationHandler, Serializable {

  private static final long serialVersionUID = 1L;

  protected static final Logger LOG = Logger.getLogger(Context.class.getName());

  // This context is used to fetch the caller id.
  // It is not used to get the API-Type. This is done by the
  // AccessorUtil context.
  private static final Context CONTEXT = Context.getFromEnvironment();

  private final Object api;

  public APIInterceptor(final Object api) {
    this.api = api;
  }

  class APIInterceptorCommand implements Command<Object> {

    private static final long serialVersionUID = 6293853276365124717L;
    private final Method m;
    private final Object[] args;
    private final boolean isSecuritySet;

    public APIInterceptorCommand(final Method m, final Object[] args, final boolean isSecuritySet) {
      this.args = args;
      this.m = m;
      this.isSecuritySet = isSecuritySet;
    }

    public Object execute(final Environment env) throws Exception {
      if (!this.isSecuritySet) {
        APIInterceptor.this.setSecurityContext();
      }
      try {
        return this.m.invoke(APIInterceptor.this.api, this.args);
      } catch (final InvocationTargetException e) {
        final Throwable t = e.getCause();
        if (t instanceof Exception) {
          throw (Exception) t;
        } else if (t instanceof Error) {
          throw (Error) t;
        } else {
          throw new BonitaInternalException("Ouch! Unexpected error: " + t, t);
        }
      }
    }

  }

  public Object invoke(final Object proxy, final Method method, final Object[] args) throws BonitaException {
    try {
      // If already in environment: use the same environment to execute the
      // command directly
      if (Environment.getCurrent() != null) {
        return new APIInterceptorCommand(method, args, true).execute(Environment.getCurrent());
      }
      // If no environment: use directly the command Service with interceptors
      final EnvironmentFactory envFactory = GlobalEnvironmentFactory.getEnvironmentFactory();
      return envFactory.get(CommandService.class).execute(new APIInterceptorCommand(method, args, false));
    } catch (final PvmException e) {
      final Throwable cause = e.getCause();
      if (cause instanceof BonitaException) {
        throw (BonitaException) cause;
      }
      throw e;
    } catch (final BonitaException e) {
      throw e;
    } catch (final InvalidEnvironmentException e) {
      // Don't throw BonitaInternalException if Environment parsing failed.
      // This is a configuration problem, not an engine bug.
      throw e;
    } catch (final BonitaWrapperException e) {
      throw (RuntimeException) e.getCause();
    } catch (final RuntimeException e) {
      throw BonitaInternalException.build(e);
    } catch (final Exception t) {
      throw new BonitaInternalException("Ouch! Unexpected error: " + t, t);
    }
  }

  private void setSecurityContext() {
    switch (APIInterceptor.CONTEXT) {
      case EJB3:
        if (APIInterceptor.LOG.isLoggable(Level.FINE)) {
          APIInterceptor.LOG.fine("EJB3 Context found, using EJBContext to get the caller ID");
        }
        try {
          final Object o = new InitialContext().lookup("java:comp/EJBContext");
          final Class< ? > ejbContextClass = Class.forName("javax.ejb.EJBContext");
          final Method getCallerPrincipalMethod = ejbContextClass.getMethod("getCallerPrincipal");
          final Principal principal = (Principal) getCallerPrincipalMethod.invoke(o);
          final String callerId = principal.getName();
          Authentication.setUserId(callerId);
          break;
        } catch (final Exception e) {
          e.printStackTrace();
          throw new BonitaRuntimeException(e);
        }
      case EJB2:
        if (APIInterceptor.LOG.isLoggable(Level.FINE)) {
          APIInterceptor.LOG.fine("EJB2 Context found, using SessionContext to get the caller ID");
        }
        // Done inside EJB implementations.
        break;
      case Standard:
        if (APIInterceptor.LOG.isLoggable(Level.FINE)) {
          APIInterceptor.LOG.fine("J2SE Context found, using SecurityContext to get the caller ID");
        }
        Authentication.setUserId(SecurityContext.getCallerId());
        break;
      default:
        Misc.badStateIfTrue(true, "Ouch! Don't know how to handle this CONTEXT: " + APIInterceptor.CONTEXT);
    }
  }
}
