/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.impl;

import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.ow2.bonita.definition.GlobalClassData;
import org.ow2.bonita.definition.PackageClassData;
import org.ow2.bonita.definition.PackageDependency;
import org.ow2.bonita.deployment.Deployer;
import org.ow2.bonita.deployment.Deployment;
import org.ow2.bonita.deployment.DeploymentRuntimeException;
import org.ow2.bonita.facade.ManagementAPI;
import org.ow2.bonita.facade.RuntimeAPI;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.PackageNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.UndeletableInstanceException;
import org.ow2.bonita.facade.exception.UndeletablePackageException;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.parsing.XpdlParserException;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.Repository;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.Tool;
import org.ow2.novabpm.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes,
 *         Pierre Vigneras
 */
public class ManagementAPIImpl implements ManagementAPI {

  protected ManagementAPIImpl() {
  }

  private static final Logger LOG = Logger.getLogger(ManagementAPIImpl.class.getName());

  public Map<String, ProcessDefinition> deploy(final Deployment deployment) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(deployment);
    try {
      return Deployer.deploy(deployment, EnvTool.getUserId());
    } catch (final DeploymentRuntimeException e) {
      throw new DeploymentException(e.getMessage(), e);
    } catch (final XpdlParserException e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public Map<String, ProcessDefinition> deploy(final URL xpdlURL, final Class< ? >[] classes) throws DeploymentException {
    return this.deploy(new Deployment(xpdlURL, Tool.getClasses(classes)));
  }

  public Map<String, ProcessDefinition> deployXpdl(final byte[] xpdlFile) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(xpdlFile);
    return this.deploy(new Deployment(xpdlFile, null));
  }

  public Map<String, ProcessDefinition> deployXpdl(final URL xpdlURL) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(xpdlURL);
    return this.deploy(new Deployment(xpdlURL, null));
  }

  public Map<String, ProcessDefinition> deployZip(final byte[] zipFile) throws DeploymentException {
    return this.deployArchive(zipFile);
  }

  public Map<String, ProcessDefinition> deployZip(final URL zipURL) throws DeploymentException {
    return this.deployArchive(zipURL);
  }

  public Map<String, ProcessDefinition> deployBar(final URL jarFile) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(jarFile);
    return this.deployArchive(jarFile);
  }

  public Map<String, ProcessDefinition> deployBar(final byte[] barFile) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(barFile);
    return this.deployArchive(barFile);
  }

  public void deployClass(final byte[] clazz) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(clazz);
    String className;
    try {
      className = Tool.visitClass(clazz).getClassName();
    } catch (final ArrayIndexOutOfBoundsException e) {
      throw new DeploymentException("The byte table is not correct parameter.");
    }
    final Repository repository = EnvTool.getRepository();
    final GlobalClassData globalClassData = repository.getGlobalClassData(className);
    if (globalClassData != null) {
      // this class was already deployed
      throw new DeploymentException("There is already a deployed class with this name: ", className);
    }
    repository.storeGlobalClassData(new GlobalClassData(className, clazz));
  }

  public void deployClassesInJar(final byte[] classesArchive) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(classesArchive);
    final File tmpDir = this.explodeArchive(classesArchive, "globalClasses");
    final Set<byte[]> classes = this.getClasses(tmpDir);
    this.deployClasses(classes);
    Misc.deleteDir(tmpDir);
  }

  public void deployClasses(final Set<byte[]> classes) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(classes);
    for (final byte[] data : classes) {
      this.deployClass(data);
    }
  }

  public void undeploy(final PackageDefinitionUUID packageUUID) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(packageUUID);
    try {
      Deployer.undeployPackage(packageUUID, EnvTool.getUserId());
    } catch (final DeploymentRuntimeException e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public void removeClass(final String className) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(className);

    final Repository repository = EnvTool.getRepository();
    if (repository.getGlobalClassData(className) == null) {
      throw new DeploymentException("There is no class defined in global class repository  with name: ", className);

    }
    // check that no process is still using this class
    final Set<PackageDependency> packageDependencies = repository.getPackageDependencies(className);
    if (packageDependencies != null && !packageDependencies.isEmpty()) {
      for (final PackageDependency packageDependency : packageDependencies) {
        final PackageClassData data = repository.getPackageClassData(packageDependency.getPackageDefinitionUUID());
        if (data == null || !data.getClasses().containsKey(className)) {
          throw new DeploymentException("Package is still using global class. " + "Cannot undeploy class.",
            className, packageDependency.getPackageDefinitionUUID());
        }
      }
    }
    repository.removeGlobalClassData(className);
    ClassDataLoader.removeGlobalClassLoader(className);
  }

  public void removeClasses(final String[] classNames) throws DeploymentException {
    if (classNames != null) {
      for (final String className : classNames) {
        this.removeClass(className);
      }
    }
  }

  public void replaceClass(final String className, final byte[] newClazz) throws DeploymentException {
    this.removeClass(className);
    this.deployClass(newClazz);
  }

  public void clearHistory() {
    final Archiver archiver = EnvTool.getArchiver();
    archiver.clear();
  }

  private Map<String, ProcessDefinition> deployArchive(final URL archiveURL) throws DeploymentException {
    InputStream stream = null;
    try {
      stream = archiveURL.openStream();
      return this.deployArchive(Misc.getAllContentFrom(stream));
    } catch (final IOException e) {
      throw new DeploymentException("Problem while deploying bar: ", e);
    } finally {
      Misc.close(stream);
    }
  }

  private Map<String, ProcessDefinition> deployArchive(final byte[] barFile) throws DeploymentException {
    File tmpDir = null;
    Map<String, ProcessDefinition> processes = null;
    try {
      tmpDir = this.explodeArchive(barFile, "tmpBonitaBar");
      final List<File> xpdlFiles = this.getMatchingFiles(tmpDir, ".xpdl");
      if (xpdlFiles.isEmpty()) {
        throw new DeploymentException("The given bar archive does not contain any '.xpdl' file");
      }
      if (xpdlFiles.size() > 1) {
        throw new DeploymentException("The given bar archive contains many '.xpdl' files");
      }
      URL xpdlURL;
      try {
        xpdlURL = xpdlFiles.get(0).toURL();
        if (ManagementAPIImpl.LOG.isLoggable(Level.FINE)) {
          ManagementAPIImpl.LOG.fine("Xpdl file currently deployed: " + xpdlURL);
        }
      } catch (final MalformedURLException e) {
        throw new DeploymentException("Error when deploying the bar file", e);
      }
      processes = this.deploy(new Deployment(xpdlURL, this.getClasses(tmpDir)));
    } finally {
      if (tmpDir != null) {
        Misc.deleteDir(tmpDir);
      }
    }
    return processes;
  }

  private String getClassName(final File classFile, final File baseDir) {
    final String basePath = baseDir.getAbsolutePath();
    final String classPath = classFile.getAbsolutePath();
    String className = classPath.substring(basePath.length() + 1);
    final String suffix = ".class";
    className = className.substring(0, className.length() - suffix.length());
    return className.replace(File.separatorChar, '.');
  }

  private List<File> getMatchingFiles(final File baseDir, final String extension) {
    final List<File> files = new ArrayList<File>();
    this.getMatchingFiles(baseDir, files, extension);
    return files;
  }

  private void getMatchingFiles(final File dir, final List<File> matchingFiles, final String extension) {
    for (final File f : dir.listFiles()) {
      if (f.getName().endsWith(extension)) {
        matchingFiles.add(f);
      } else if (f.isDirectory()) {
        this.getMatchingFiles(f, matchingFiles, extension);
      }
    }
  }

  private File explodeArchive(final byte[] archive, final String destDirName) throws DeploymentException {
    File tmpDir = null;
    try {
      // Create temp dir
      new File(System.getProperty("java.io.tmpdir")).mkdirs();
      tmpDir = Tool.createTempFile(destDirName, null, null);
      // tmpDir is an empty file at this stage, we want a directory! 
      Misc.badStateIfFalse(tmpDir.delete(), "Can't delete: " + tmpDir);
      if (!tmpDir.mkdirs()) {
        throw new DeploymentException("Cannot create the temporary directory:' " + tmpDir);
      }

      // Unzip file
      final InputStream in = new ByteArrayInputStream(archive);
      ZipInputStream zis = null;
      zis = new ZipInputStream(in);

      ZipEntry zipEntry = null;
      while ((zipEntry = zis.getNextEntry()) != null) {
        // write the file to the disk
        // Create file
        final File outputFile = new File(tmpDir.getAbsolutePath() + File.separator + zipEntry.getName());
        if (zipEntry.isDirectory()) {
          outputFile.mkdirs();
        } else {
          int i = 0;
          int position = 0;
          while (i < zipEntry.getName().length()) {
            if (zipEntry.getName().charAt(i) == File.separatorChar) {
              position = i + 1;
            }
            i++;
          }
          new File(tmpDir.getAbsolutePath() + File.separator + zipEntry.getName().substring(0, position)).mkdirs();
          final BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(outputFile));
          int c;
          final byte[] buffer = new byte[512];
          while ((c = zis.read(buffer)) != -1) {
            bos.write(buffer, 0, c);
          }
          // bos.flush();
          Misc.close(bos);
        }
      }
      zis.close();
      in.close();
    } catch (final IOException e) {
      throw new DeploymentException("Problem while exploding jar file at deployment", e);
    }
    return tmpDir;
  }

  private Set<byte[]> getClasses(final File tmpDir) throws DeploymentException {
    if (ManagementAPIImpl.LOG.isLoggable(Level.FINE)) {
      ManagementAPIImpl.LOG.fine("Looking for datas in directory: " + tmpDir);
    }
    final List<File> classFiles = this.getMatchingFiles(tmpDir, ".class");
    final Set<byte[]> classes = new HashSet<byte[]>();
    try {
      for (final File classFile : classFiles) {
        final String className = this.getClassName(classFile, tmpDir);
        final URL classURL = classFile.toURL();
        if (ManagementAPIImpl.LOG.isLoggable(Level.FINE)) {
          ManagementAPIImpl.LOG.fine("Found class: " + className + " with URL: " + classURL);
        }
        classes.add(Misc.getAllContentFrom(classURL));
      }
    } catch (final IOException e) {
      throw new DeploymentException("Error when getting classDatas at deployment ", e);
    }
    return classes;
  }

  @SuppressWarnings("null")
  public void deletePackage(final PackageDefinitionUUID packageDefinitionUUID)
    throws PackageNotFoundException, UndeletablePackageException, UndeletableInstanceException {
    FacadeUtil.checkArgsNotNull(packageDefinitionUUID);

    final Querier journal = EnvTool.getJournalQueriers();
    final Querier history = EnvTool.getHistoryQueriers();
    final Repository repository = EnvTool.getRepository();

    final PackageFullDefinition journalPack = journal.getPackage(packageDefinitionUUID);
    final boolean inJournal = journalPack != null;
    final PackageFullDefinition historyPack = history.getPackage(packageDefinitionUUID);
    final boolean inHistory = historyPack != null;

    if (!inJournal && !inHistory) {
      throw new PackageNotFoundException(packageDefinitionUUID);
    }

    Set<ProcessFullDefinition> processes;
    String packageId;
    String packageVersion;
    if (inJournal) {
      processes = journalPack.getProcesses();
      packageId = journalPack.getPackageId();
      packageVersion = journalPack.getVersion();
    } else {
      processes = historyPack.getProcesses();
      packageId = historyPack.getPackageId();
      packageVersion = historyPack.getVersion();
    }

    repository.removePackageVersion(packageId, packageVersion);
    final RuntimeAPI runtimeAPI = new RuntimeAPIImpl();
    try {
      for (final ProcessFullDefinition process : processes) {
        final ProcessDefinitionUUID processUUID = process.getUUID();
        runtimeAPI.deleteAllProcessInstances(processUUID);
        repository.removeProcessVersion(process.getProcessId(), process.getVersion(), packageId);
      }
    } catch (final ProcessNotFoundException e) {
      throw new BonitaRuntimeException(e);
    }

    if (inJournal) {
      // check that none of the processes have running instances
      for (final ProcessFullDefinition process : processes) {
        final ProcessDefinitionUUID processUUID = process.getUUID();
        final Set<XpdlInstance> xpdlInstances = repository.getXpdlInstances(processUUID);
        if (!xpdlInstances.isEmpty()) {
          final ProcessInstanceUUID instanceUUID = xpdlInstances.iterator().next().getUUID();
          throw new UndeletablePackageException(packageDefinitionUUID, processUUID, instanceUUID);
        }

      }
      Deployer.removePackageDependencies(journalPack);
      final Recorder recorder = EnvTool.getRecorder();
      recorder.remove(journalPack);
    }

    if (inHistory) {
      final Archiver archiver = EnvTool.getArchiver();
      archiver.remove(historyPack);
    }

  }
}
