/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.impl;

import java.util.HashSet;
import java.util.Set;

import org.ow2.bonita.facade.QueryDefinitionAPI;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.def.majorElement.impl.ActivityDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.DataFieldDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.PackageDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.ParticipantDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.ProcessDefinitionImpl;
import org.ow2.bonita.facade.exception.ActivityDefNotFoundException;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.DataFieldNotFoundException;
import org.ow2.bonita.facade.exception.PackageNotFoundException;
import org.ow2.bonita.facade.exception.ParticipantNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ParticipantDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.util.EnvTool;


public class QueryDefinitionAPIImpl implements QueryDefinitionAPI {

  protected QueryDefinitionAPIImpl() {
  }
  
  private Set<ProcessDefinition> getProcessCopy(Set<ProcessDefinition> src) {
    Set<ProcessDefinition> result = new HashSet<ProcessDefinition>();
    if (src != null) {
      for (ProcessDefinition p : src) {
        result.add(new ProcessDefinitionImpl(p));
      }
    }
    return result;
  }

  private Set<PackageDefinition> getPackageCopy(Set< PackageDefinition> src) {
    Set<PackageDefinition> result = new HashSet<PackageDefinition>();
    if (src != null) {
      for (PackageDefinition p : src) {
        result.add(new PackageDefinitionImpl(p));
      }
    }
    return result;
  }

  private Set<ActivityDefinition> getActivityCopy(Set<ActivityDefinition> src) {
    Set<ActivityDefinition> result = new HashSet<ActivityDefinition>();
    if (src != null) {
      for (ActivityDefinition p : src) {
        result.add(new ActivityDefinitionImpl(p));
      }
    }
    return result;
  }

  private Set<ParticipantDefinition> getParticipantCopy(Set<ParticipantDefinition> src) {
    Set<ParticipantDefinition> result = new HashSet<ParticipantDefinition>();
    if (src != null) {
      for (ParticipantDefinition p : src) {
        result.add(new ParticipantDefinitionImpl(p));
      }
    }
    return result;
  }

  private Set<DataFieldDefinition> getDataFieldCopy(Set<DataFieldDefinition> src) {
    Set<DataFieldDefinition> result = new HashSet<DataFieldDefinition>();
    if (src != null) {
      for (DataFieldDefinition p : src) {
        result.add(new DataFieldDefinitionImpl(p));
      }
    }
    return result;
  }

  private <T> void add(T src, Set<T> dest) {
    if (src != null) {
      dest.add(src);
    }
  }

  private <T> void add(Set<T> src, Set< ? super T> dest) {
    if (src != null) {
      for (T t : src) {
        add(t, dest);
      }
    }
  }

  /*
   * API
   */

  public PackageDefinition getLastPackage(String packageId) throws PackageNotFoundException {
    FacadeUtil.checkArgsNotNull(packageId);
    PackageFullDefinition last = EnvTool.getAllQueriers().getLastDeployedPackage(packageId, 
        PackageDefinition.PackageState.DEPLOYED);
    if (last == null) {
      throw new PackageNotFoundException(packageId);
    }
    return new PackageDefinitionImpl(last);
  }

  public ProcessDefinition getLastProcess(String packageId, String processId) throws ProcessNotFoundException {
    FacadeUtil.checkArgsNotNull(processId);
    ProcessFullDefinition last = EnvTool.getAllQueriers().getLastDeployedProcess(packageId, processId, 
        ProcessDefinition.ProcessState.DEPLOYED);
    if (last == null) {
      throw new ProcessNotFoundException(processId);
    }
    return new ProcessDefinitionImpl(last);
  }

  public PackageDefinition getPackage(PackageDefinitionUUID packageUUID) throws PackageNotFoundException {
    FacadeUtil.checkArgsNotNull(packageUUID);
    PackageFullDefinition package1 = EnvTool.getAllQueriers().getPackage(packageUUID);
    if (package1 == null) {
      throw new PackageNotFoundException(packageUUID);
    }
    return new PackageDefinitionImpl(package1);
  }
  
  public PackageDefinition getPackage(String packageId, String version) throws PackageNotFoundException {
    FacadeUtil.checkArgsNotNull(packageId, version);
    PackageFullDefinition package1 = EnvTool.getAllQueriers().getPackage(packageId, version);
    if (package1 == null) {
      throw new PackageNotFoundException(packageId, version);
    }
    return new PackageDefinitionImpl(package1);
  }

  @SuppressWarnings("unchecked")
  public Set<ProcessDefinition> getPackageProcesses(PackageDefinitionUUID packageUUID) throws PackageNotFoundException {
    FacadeUtil.checkArgsNotNull(packageUUID);
    Set<ProcessFullDefinition> processes = EnvTool.getAllQueriers().getPackageProcesses(packageUUID);
    if (processes == null || processes.isEmpty()) {
      // check that package exists (throws exception)
      getPackage(packageUUID);
    }
    return getProcessCopy((Set)processes);
  }

  @SuppressWarnings("unchecked")
  public Set<PackageDefinition> getPackages() {
    Set<PackageFullDefinition> packages = EnvTool.getAllQueriers().getPackages();
    return getPackageCopy((Set)packages);
  }

  @SuppressWarnings("unchecked")
  public Set<PackageDefinition> getPackages(String packageId) {
    FacadeUtil.checkArgsNotNull(packageId);
    Set<PackageFullDefinition> packages = EnvTool.getAllQueriers().getPackages(packageId);
    return getPackageCopy((Set)packages);
  }

  public ProcessDefinition getProcess(ProcessDefinitionUUID processUUID) throws ProcessNotFoundException {
    FacadeUtil.checkArgsNotNull(processUUID);
    ProcessFullDefinition process = EnvTool.getAllQueriers().getProcess(processUUID);
    if (process == null) {
      throw new ProcessNotFoundException(processUUID);
    }
    return new ProcessDefinitionImpl(process);
  }

  @SuppressWarnings("unchecked")
  public Set<ProcessDefinition> getProcesses(String processId, String processVersion) {
    FacadeUtil.checkArgsNotNull(processId, processVersion);
    Set<ProcessFullDefinition> processes = EnvTool.getAllQueriers().getProcesses(processId, processVersion);
    return getProcessCopy((Set)processes);
  }
  
  public ProcessDefinition getProcess(String packageId, String processId, String processVersion) throws ProcessNotFoundException {
    FacadeUtil.checkArgsNotNull(packageId, processId, processVersion);
    ProcessFullDefinition process = EnvTool.getAllQueriers().getProcess(packageId, processId, processVersion);
    if (process == null) {
      throw new ProcessNotFoundException(processId, processVersion);
    }
    return new ProcessDefinitionImpl(process);
  }

  public Set<ActivityDefinition> getProcessActivities(ProcessDefinitionUUID processUUID) throws ProcessNotFoundException {
    FacadeUtil.checkArgsNotNull(processUUID);
    ProcessFullDefinition process = EnvTool.getAllQueriers().getProcess(processUUID);
    if (process == null) {
      throw new ProcessNotFoundException(processUUID);
    }
    Set<ActivityDefinition> activities = process.getActivities();
    return getActivityCopy(activities);
  }

  public ActivityDefinition getProcessActivity(ProcessDefinitionUUID processUUID, String activityId)
    throws ProcessNotFoundException, ActivityNotFoundException {
    FacadeUtil.checkArgsNotNull(processUUID, activityId);
    ProcessFullDefinition process = EnvTool.getAllQueriers().getProcess(processUUID);
    if (process == null) {
      throw new ProcessNotFoundException(processUUID);
    }
    Set<ActivityDefinition> activities = process.getActivities();
    if (activities != null) {
      for (ActivityDefinition activityDefinition : activities) {
        if (activityDefinition.getActivityId().equals(activityId)) {
          return new ActivityDefinitionImpl(activityDefinition);
        }
      }
    }
    throw new ActivityNotFoundException(processUUID, activityId);
  }

  public ParticipantDefinition getProcessParticipant(ProcessDefinitionUUID processUUID, String participantId)
    throws ProcessNotFoundException, ParticipantNotFoundException {
    FacadeUtil.checkArgsNotNull(processUUID, participantId);
    Set<ParticipantDefinition> participants = getProcessParticipants(processUUID);
    if (participants != null) {
      for (ParticipantDefinition participant : participants) {
        if (participant.getParticipantId().equals(participantId)) {
          return new ParticipantDefinitionImpl(participant);
        }
      }
    }
    throw new ParticipantNotFoundException(participantId, processUUID);
  }

  public Set<ParticipantDefinition> getProcessParticipants(ProcessDefinitionUUID processUUID) throws ProcessNotFoundException {
    FacadeUtil.checkArgsNotNull(processUUID);
    ProcessFullDefinition process = EnvTool.getAllQueriers().getProcess(processUUID);
    if (process == null) {
      throw new ProcessNotFoundException(processUUID);
    }
    Set<ParticipantDefinition> participants = process.getParticipants();
    return getParticipantCopy(participants);
  }

  @SuppressWarnings("unchecked")
  public Set<ProcessDefinition> getProcesses() {
    Set<ProcessFullDefinition> processes = EnvTool.getAllQueriers().getProcesses();
    return getProcessCopy((Set)processes);
  }

  @SuppressWarnings("unchecked")
  public Set<ProcessDefinition> getProcesses(String processId) {
    FacadeUtil.checkArgsNotNull(processId);
    Set<ProcessFullDefinition> processes = EnvTool.getAllQueriers().getProcesses(processId);
    return getProcessCopy((Set)processes);
  }
  
  @SuppressWarnings("unchecked")
  public Set<ProcessDefinition> getPackageProcesses(String packageId, String processId) {
    FacadeUtil.checkArgsNotNull(packageId, processId);
    Set<ProcessFullDefinition> processes = EnvTool.getAllQueriers().getPackageProcesses(packageId, processId);
    return getProcessCopy((Set)processes);
  }

  public ParticipantDefinitionUUID getProcessParticipantId(ProcessDefinitionUUID processUUID, String participantName)
    throws ProcessNotFoundException {
    FacadeUtil.checkArgsNotNull(processUUID, participantName);
    Set<ParticipantDefinition> participants = getProcessParticipants(processUUID);
    if (participants != null) {
      for (ParticipantDefinition participant : participants) {
        if (participantName.equals(participant.getName())) {
          return participant.getUUID();
        }
      }
    }
    return null;
  }

  public ActivityDefinitionUUID getProcessActivityId(ProcessDefinitionUUID processUUID, String activityName)
    throws ProcessNotFoundException {
    FacadeUtil.checkArgsNotNull(processUUID, activityName);
    Set<ActivityDefinition> activities = getProcessActivities(processUUID);
    if (activities != null) {
      for (ActivityDefinition activity : activities) {
        if (activityName.equals(activity.getName())) {
          return activity.getUUID();
        }
      }
    }
    return null;
  }

  @SuppressWarnings("unchecked")
  public Set<PackageDefinition> getPackages(String packageId, PackageDefinition.PackageState packageState) {
    FacadeUtil.checkArgsNotNull(packageId, packageState);
    Set<PackageFullDefinition> packages = EnvTool.getAllQueriers().getPackages(packageId, packageState);
    return getPackageCopy((Set)packages);
  }
  @SuppressWarnings("unchecked")
  public Set<PackageDefinition> getPackages(PackageDefinition.PackageState packageState) {
    FacadeUtil.checkArgsNotNull(packageState);
    Set<PackageFullDefinition> packages = EnvTool.getAllQueriers().getPackages(packageState);
    return getPackageCopy((Set)packages);
  }
  @SuppressWarnings("unchecked")
  public Set<ProcessDefinition> getProcesses(String processId, ProcessDefinition.ProcessState processState) {
    FacadeUtil.checkArgsNotNull(processId, processId);
    Set<ProcessFullDefinition> processes = EnvTool.getAllQueriers().getProcesses(processId, processState);
    return getProcessCopy((Set)processes);
  }

  @SuppressWarnings("unchecked")
  public Set<ProcessDefinition> getProcesses(ProcessDefinition.ProcessState processState) {
    FacadeUtil.checkArgsNotNull(processState);
    Set<ProcessFullDefinition> processes = EnvTool.getAllQueriers().getProcesses(processState);
    return getProcessCopy((Set)processes);
  }

  public ProcessDefinition getPackageProcess(PackageDefinitionUUID packageDefinitionUUID, String processId)
    throws PackageNotFoundException, ProcessNotFoundException {
    FacadeUtil.checkArgsNotNull(packageDefinitionUUID, processId);
    ProcessDefinition process = EnvTool.getAllQueriers().getPackageProcess(packageDefinitionUUID, processId);
    if (process == null) {
      if (EnvTool.getAllQueriers().getPackage(packageDefinitionUUID) == null) {
        throw new PackageNotFoundException(packageDefinitionUUID);
      }
      throw new ProcessNotFoundException(processId, packageDefinitionUUID);
    }
    return new ProcessDefinitionImpl(process);
  }

  public DataFieldDefinition getProcessDataField(ProcessDefinitionUUID processDefinitionUUID, String dataFieldId)
    throws ProcessNotFoundException, DataFieldNotFoundException {
    FacadeUtil.checkArgsNotNull(processDefinitionUUID, dataFieldId);
    Set<DataFieldDefinition> datafields = getProcessDataFields(processDefinitionUUID);
    if (datafields != null) {
      for (DataFieldDefinition datefield : datafields) {
        if (datefield.getDataFieldId().equals(dataFieldId)) {
          return new DataFieldDefinitionImpl(datefield);
        }
      }
    }
    throw new DataFieldNotFoundException(dataFieldId, processDefinitionUUID);
  }

  public Set<DataFieldDefinition> getProcessDataFields(ProcessDefinitionUUID processDefinitionUUID)
    throws ProcessNotFoundException {
    FacadeUtil.checkArgsNotNull(processDefinitionUUID);
    ProcessFullDefinition process = EnvTool.getAllQueriers().getProcess(processDefinitionUUID);
    if (process == null) {
      throw new ProcessNotFoundException(processDefinitionUUID);
    }
    Set<DataFieldDefinition> datafields = process.getDataFields();
    return getDataFieldCopy(datafields);
  }

  public Set<DataFieldDefinition> getActivityDataFields(ActivityDefinitionUUID activityDefinitionUUID)
    throws ActivityDefNotFoundException {
    FacadeUtil.checkArgsNotNull(activityDefinitionUUID);
    ActivityDefinition activity = EnvTool.getAllQueriers().getActivity(activityDefinitionUUID);
    if (activity == null) {
      throw new ActivityDefNotFoundException(activityDefinitionUUID);
    }
    Set<DataFieldDefinition> datafields = activity.getDataFields();
    return getDataFieldCopy(datafields);
  }

  public DataFieldDefinition getActivityDataField(ActivityDefinitionUUID activityDefinitionUUID, String dataFieldId)
    throws ActivityDefNotFoundException, DataFieldNotFoundException {
    FacadeUtil.checkArgsNotNull(activityDefinitionUUID);
    ActivityDefinition activity = EnvTool.getAllQueriers().getActivity(activityDefinitionUUID);
    if (activity == null) {
      throw new ActivityDefNotFoundException(activityDefinitionUUID);
    }
    Set<DataFieldDefinition> datafields = activity.getDataFields();
    if (datafields != null) {
      for (DataFieldDefinition datefield : datafields) {
        if (datefield.getDataFieldId().equals(dataFieldId)) {
          return new DataFieldDefinitionImpl(datefield);
        }
      }
    }
    throw new DataFieldNotFoundException(dataFieldId, activityDefinitionUUID);
  }

}

