package org.ow2.bonita.persistence.db;

import java.util.HashSet;
import java.util.Set;

import org.jbpm.pvm.Execution;
import org.jbpm.pvm.job.Timer;
import org.ow2.bonita.definition.GlobalClassData;
import org.ow2.bonita.definition.PackageClassData;
import org.ow2.bonita.definition.PackageDependency;
import org.ow2.bonita.definition.XpdlProcess;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.persistence.RuntimeDbSession;
import org.ow2.bonita.runtime.ActivityInstanceKey;
import org.ow2.bonita.runtime.PackageVersion;
import org.ow2.bonita.runtime.ProcessInstanceKey;
import org.ow2.bonita.runtime.ProcessVersion;
import org.ow2.bonita.runtime.Repository;
import org.ow2.bonita.runtime.TaskRunTime;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.util.EnvTool;
import org.ow2.novabpm.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class DbRepository implements Repository {

  private String persistenceServiceName;
  private DbRepositoryBuffer buffer = new DbRepositoryBuffer();
  
  private RuntimeDbSession getRuntimeDbSession() {
    return EnvTool.getRuntimeDbSession(persistenceServiceName);
  }

  public DbRepository(String persistenceServiceName) {
    this.persistenceServiceName = persistenceServiceName;
  }

  //INSTANCE REPOSITORY
  public XpdlInstance getXpdlInstance(ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    if (buffer.isInstanceRemoved(instanceUUID)) {
      return null;
    }
    XpdlInstance instance = buffer.getInstance(instanceUUID);
    if (instance != null) {
      return instance;
    }
    instance = getRuntimeDbSession().getXpdlInstance(instanceUUID);
    if (instance != null) {
      buffer.addInstance(instance);
    }
    return instance;
  }

  public Set<XpdlInstance> getXpdlInstances(ProcessDefinitionUUID processUUID) {
    Misc.checkArgsNotNull(processUUID);
    Set<XpdlInstance> dbInstances = getRuntimeDbSession().getXpdlInstances(processUUID);
    for (XpdlInstance instance : dbInstances) {
      if (!buffer.containsInstance(instance.getUUID()) && !buffer.isInstanceRemoved(instance.getUUID())) {
        buffer.addInstance(instance);
      }
    }
    Set<XpdlInstance> result = new HashSet<XpdlInstance>();
    for (XpdlInstance instance : buffer.getInstances().values()) {
      if (instance.getProcessDefinitionUUID().equals(processUUID)) {
        result.add(instance);
      }
    }
    return result;
  }

  public void storeXpdlInstance(XpdlInstance instance) {
    Misc.checkArgsNotNull(instance);
    getRuntimeDbSession().save(instance);
    buffer.addInstance(instance);
  }

  public XpdlInstance removeXpdlInstance(ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    XpdlInstance instance = getXpdlInstance(instanceUUID);
    if (instance == null) {
      return null;
    }
    buffer.removeInstance(instance);
    getRuntimeDbSession().delete(instance);
    return instance;
  }

  private XpdlExecution getExecOnNode(XpdlExecution exec, String nodeName, String iterationId, String activityInstanceId) {
    Misc.checkArgsNotNull(exec, nodeName, iterationId, activityInstanceId);
    if (exec.getExecutions() == null || exec.getExecutions().isEmpty()) {
      if (exec.getNode() != null 
          && exec.getNode().getName().equals(nodeName)
          && exec.getIterationId().equals(iterationId)
          && exec.getActivityInstanceId().equals(activityInstanceId)) {
        return exec;
      }
    }
    for (Execution child : exec.getExecutions()) {
      XpdlExecution found = getExecOnNode((XpdlExecution) child, nodeName, iterationId, activityInstanceId);
      if (found != null) {
        return found;
      }
    }
    return null;
  }

  public XpdlExecution getExecutionPointingOnNode(
      ProcessInstanceUUID instanceUUID, String nodeName, String iterationId, String activityInstanceId) {
    Misc.checkArgsNotNull(instanceUUID, nodeName, iterationId, activityInstanceId);
    
    XpdlInstance instance = buffer.getInstance(instanceUUID);
    if (instance != null) {
      return getExecOnNode(instance.getRootExecution(), nodeName, iterationId, activityInstanceId);
    }
    XpdlExecution exec = 
      getRuntimeDbSession().getExecutionPointingOnNode(instanceUUID, nodeName, iterationId, activityInstanceId);
    if (exec != null && !buffer.isInstanceRemoved(exec.getXpdlInstance().getUUID())) {
      return exec;
    }
    return null;
  }

  //PROCESS REPOSITORY
  public XpdlProcess getXpdlProcess(ProcessDefinitionUUID processUUID) {
    Misc.checkArgsNotNull(processUUID);
    
    if (buffer.isProcessRemoved(processUUID)) {
      return null;
    }
    XpdlProcess process = buffer.getProcess(processUUID);
    if (process != null) {
      return process;
    }
    process = getRuntimeDbSession().getXpdlProcess(processUUID);
    if (process != null) {
      buffer.addProcess(process);
    }
    return process;
  }

  public void storeXpdlProcess(XpdlProcess process) {
    Misc.checkArgsNotNull(process);
    buffer.addProcess(process);
    getRuntimeDbSession().save(process);
  }
  
  public String getLastProcessVersion(String processId, String packageId) {
    Misc.checkArgsNotNull(processId, packageId);
    ProcessVersion pv = getRuntimeDbSession().getLastProcessVersion(processId, packageId);
    if (pv != null) {
      return pv.getProcessVersion();
    }
    return "";
  }

  public void storeProcessVersion(String processId, String processVersion, String packageId) {
    Misc.checkArgsNotNull(processId, processVersion, packageId);
    ProcessVersion pv = getRuntimeDbSession().getLastProcessVersion(processId, packageId);
    if (pv == null) {
      pv = new ProcessVersion(processId, processVersion, packageId);
      getRuntimeDbSession().save(pv);
    } else {
      pv.setProcessVersion(processVersion);
    }
  }
  
  public void removeProcessVersion(String processId, String processVersion, String packageId) {
    Misc.checkArgsNotNull(processId, processVersion, packageId);
    ProcessVersion pv = getRuntimeDbSession().getLastProcessVersion(processId, packageId);
    if (pv != null) {
      boolean stillExist = pv.removeVersion(processVersion);
      if (!stillExist) {
        getRuntimeDbSession().delete(pv);
      }
    }
  }

  public String getLastPackageVersion(String packageId) {
    PackageVersion pv = getRuntimeDbSession().getLastPackageVersion(packageId);
    if (pv != null) {
      return pv.getPackageVersion();
    }
    return "";
  }
  public void storePackageVersion(String packageId, String packageVersion) {
    PackageVersion pv = getRuntimeDbSession().getLastPackageVersion(packageId);
    if (pv == null) {
      pv = new PackageVersion(packageId, packageVersion);
      getRuntimeDbSession().save(pv);
    } else {
      pv.setPackageVersion(packageVersion);
    }
  }
  public void removePackageVersion(String packageId, String packageVersion) {
    PackageVersion pv = getRuntimeDbSession().getLastPackageVersion(packageId);
    if (pv != null) {
      boolean stillExist = pv.removeVersion(packageVersion);
      if (!stillExist) {
        getRuntimeDbSession().delete(pv);
      }
    }
  }

  public XpdlProcess removeXpdlProcess(ProcessDefinitionUUID processUUID) {
    XpdlProcess p = getXpdlProcess(processUUID);
    if (p == null) {
      return null;
    }
    getRuntimeDbSession().delete(p);
    return p;
  }


  //TASK REPOSITORY
  public void addTask(TaskRunTime task) {
    Misc.checkArgsNotNull(task);
    buffer.addTask(task);
    getRuntimeDbSession().save(task);
  }

  public TaskRunTime getTask(TaskUUID taskUUID) {
    Misc.checkArgsNotNull(taskUUID);
    
    if (buffer.isTaskRemoved(taskUUID)) {
      return null;
    }
    TaskRunTime task = buffer.getTask(taskUUID);
    if (task != null) {
      return task;
    }
    task = getRuntimeDbSession().getTask(taskUUID);
    if (task != null) {
      buffer.addTask(task);
    }
    return task;
  }
  
  public Set<TaskRunTime> getInstanceTasks(ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    if (buffer.isInstanceRemoved(instanceUUID)) {
      return new HashSet<TaskRunTime>();
    }
    
    Set<TaskRunTime> dbTasks = getRuntimeDbSession().getInstanceTasks(instanceUUID);
    for (TaskRunTime task : dbTasks) {
      if (!buffer.isTaskRemoved(task.getUUID())) {
        buffer.addTask(task);
      }
    }
    
    Set<TaskRunTime> result = new HashSet<TaskRunTime>();
    for (TaskRunTime task : buffer.getTasks().values()) {
      if (task.getInstanceUUID().equals(instanceUUID)) {
        result.add(task);
      }
    }
    
    return result;
  }

  public void removeTask(TaskRunTime task) {
    Misc.checkArgsNotNull(task);
    buffer.removeTask(task);
    getRuntimeDbSession().delete(task);
  }

  //PACKAGE CLASS DATA
  public PackageClassData getPackageClassData(PackageDefinitionUUID packageUUID) {
    Misc.checkArgsNotNull(packageUUID);
    
    if (buffer.isPackageClassDataRemoved(packageUUID)) {
      return null;
    }
    PackageClassData packCD = buffer.getPackageClassData(packageUUID);
    if (packCD != null) {
      return packCD;
    }
    packCD = getRuntimeDbSession().getPackageClassData(packageUUID);
    if (packCD != null) {
      buffer.addPackageClassData(packCD);
    }
    return packCD;
  }
  public void removePackageClassData(PackageDefinitionUUID packageUUID) {
    Misc.checkArgsNotNull(packageUUID);
    PackageClassData packCD = getPackageClassData(packageUUID);
    if (packCD == null) {
      return;
    }
    buffer.removePackageClassData(packCD);
    getRuntimeDbSession().delete(packCD);
  }
  public void storePackageClassData(PackageClassData packageCD) {
    Misc.checkArgsNotNull(packageCD);
    getRuntimeDbSession().save(packageCD);
    buffer.addPackageClassData(packageCD);
  }
  
  //GLOBAL CLASS DATAS
  public GlobalClassData getGlobalClassData(String className) {
    Misc.checkArgsNotNull(className);
    if (buffer.isGlobalClassDataRemoved(className)) {
      return null;
    }
    
    GlobalClassData globalCD = buffer.getGlobalClassData(className);
    if (globalCD != null) {
      return globalCD;
    }
    globalCD = getRuntimeDbSession().getGlobalClassData(className);
    if (globalCD != null) {
      buffer.addGlobalClassData(globalCD);
    }
    return globalCD;
  }

  public void removeGlobalClassData(String className) {
    Misc.checkArgsNotNull(className);
    GlobalClassData gcd = getGlobalClassData(className);
    if (gcd == null) {
      return;
    }
    buffer.removeGlobalClassData(gcd);
    getRuntimeDbSession().delete(gcd);
  }
  
  public void storeGlobalClassData(GlobalClassData globalClassData) {
    Misc.checkArgsNotNull(globalClassData);
    getRuntimeDbSession().save(globalClassData);
    buffer.addGlobalClassData(globalClassData);
  }


  //DEPENDENT PACKAGES
  public Set<PackageDependency> getPackageDependencies(String className) {
    Misc.checkArgsNotNull(className);
    return getRuntimeDbSession().getPackageDependencies(className);
  }
  public void addPackageDependencies(PackageDefinitionUUID packageUUID, Set<String> classNames) {
    Misc.checkArgsNotNull(packageUUID);
    getRuntimeDbSession().save(new PackageDependency(packageUUID, classNames));
  }
  public PackageDependency getPackageDependency(PackageDefinitionUUID packageUUID) {
    Misc.checkArgsNotNull(packageUUID);
    return getRuntimeDbSession().getPackageDependency(packageUUID);
  }
  public void removePackageDependencies(PackageDefinitionUUID packageUUID) {
    Misc.checkArgsNotNull(packageUUID);
    PackageDependency packageDependency = getPackageDependency(packageUUID);
    if (packageDependency == null) {
      return;
    }
    getRuntimeDbSession().delete(packageDependency);
  }

  public XpdlProcess findLatestProcessById(String subProcessId) {
    return getRuntimeDbSession().findLatestProcessById(subProcessId);
  }
  public Set<Timer> getInstanceTimers(ProcessInstanceUUID instanceUUID) {
    return getRuntimeDbSession().getInstanceTimers(instanceUUID);
  }
  public void removeTimer(Timer timer) {
    EnvTool.getTimerSession().cancel(timer);
  }

  public synchronized long getNextActivityInstanceNb(ProcessInstanceUUID instanceUUID,
      String activityName) {
    ActivityInstanceKey nextPik = new ActivityInstanceKey(instanceUUID, activityName);
    getRuntimeDbSession().save(nextPik);
    getRuntimeDbSession().delete(nextPik);
    return nextPik.getNumber();
  }

  public synchronized long getNextProcessInstanceNb(ProcessDefinitionUUID processUUID) {
    ProcessInstanceKey nextPik = new ProcessInstanceKey(processUUID);
    getRuntimeDbSession().save(nextPik);
    getRuntimeDbSession().delete(nextPik);
    return nextPik.getNumber();
  }

}
