/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.persistence.log;

import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.definition.activity.AbstractActivity;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.TaskRunTime;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.services.Recorder;

/**
 * @author Pierre Vigneras
 * @date Jan 31, 2008
 */
public class LoggerRecorder implements Recorder {
  private static final Logger LOG = Logger.getLogger(LoggerRecorder.class.getName());
  private final Level level;

  public LoggerRecorder() {
    this(Level.FINE);
  }

  public LoggerRecorder(String levelName) {
    this(Level.parse(levelName));
  }

  public LoggerRecorder(Level level) {
    this.level = level;
  }

  public void recordEnterActivity(AbstractActivity activity, XpdlExecution execution) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Activity ready: " + activity.getActivityId() + " (" + execution + ")");
    }
  }
  public void recordBodyEnded(AbstractActivity activity, XpdlExecution execution) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Activity body ended: " + activity.getActivityId() + " (" + execution + ")");
    }
  }
  public void recordBodyStarted(AbstractActivity activity, XpdlExecution execution) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Activity body started: " + activity.getActivityId() + " (" + execution + ")");
    }
  }

  public void recordInstanceEnded(XpdlInstance xpdlInstance, String userId) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Ended instance: " + xpdlInstance + ", userId = " + userId);
    }
  }

  public void recordInstanceStarted(XpdlInstance xpdlInstance, String userId) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Started instance: " + xpdlInstance + ", userId = " + userId);
    }
  }

  public void recordTaskReady(TaskRunTime task) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Task ready " + task.getUUID());
    }
  }

  public void recordTaskFinished(TaskRunTime task, String userId) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Task " + task.getUUID() + " finished by " + userId);
    }
  }

  public void recordTaskStarted(TaskRunTime task, String userId) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Task " + task.getUUID() + " started by " + userId);
    }
  }
  public void recordTaskResumed(TaskRunTime task, String userId) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Task " + task.getUUID() + " resumed by " + userId + ".");
    }
  }
  public void recordTaskSuspended(TaskRunTime task, String userId) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Task " + task.getUUID() + " suspended by " + userId);
    }
  }

  public void recordPackageDeployed(PackageFullDefinition pack) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Deployed package: packageDefinitionUUID: " + pack.getPackageDefinitionUUID()
          + ", packageDefinitionUUID: " + pack.getPackageId() + ", userId: "
          + pack.getDeployedBy());
    }
  }

  public void recordPackageUndeployed(PackageFullDefinition pack, String userId) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Undeployed package: packageDefinitionUUID: " + pack.getPackageDefinitionUUID() + ", userId = " + userId);
    }
  }

  public void recordInstanceVariableUpdated(String variableId, Object variableValue,
      ProcessInstanceUUID instanceUUID, String userId) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Global variable updated: processUUID: "
          + instanceUUID + ", variableId: " + variableId + ", userId = " + userId);
    }
  }

  public void recordActivityVariableUpdated(String variableId,
      Object variableValue, ProcessInstanceUUID instanceUUID, String activityId, String iterationId, String activityInstanceId,
        String userId) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Local variable updated: processUUID: "
          + instanceUUID + ", activityId: " + activityId + ", variableId: " + variableId + ", userId = " + userId);
    }
  }

  public void recordTaskAssigned(TaskRunTime taskRunTime, String assignerUserId, Set<String> candidates, String userId) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Task " + taskRunTime.getUUID() + " Assigned by: " + assignerUserId + " Assigned to: " + userId
          + "and candidates: " + candidates);
    }
  }
  
  public void remove(PackageFullDefinition package1) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "PackageDefinition " + package1.getUUID() + " removed.");
    }
  }
  public void remove(ProcessFullInstance processInst) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "ProcessInstance " + processInst.getUUID() + " removed.");
    }
  }
  public void clear() {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Clearing recorder");
    }
  }
}
