/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.runtime;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.util.BonitaRuntimeException;

public class BonitaClassLoader extends ClassLoader {

  private static final Logger LOG = Logger.getLogger(BonitaClassLoader.class.getName());

  protected Map<String, Class< ? >> loadedClasses;
  protected Map<String, byte[]> classDatas;

  private BonitaClassLoader() {
    super(BonitaClassLoader.class.getClassLoader());
  }

  public BonitaClassLoader(Map<String, byte[]> classDatas) {
    this();
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Creating a new BonitaClassLoader... classDatas = " + classDatas);
    }
    this.classDatas = classDatas;
    if (classDatas != null) {
      for (Map.Entry<String, byte[]> classData : classDatas.entrySet()) {
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("BonitaClassLoader constructor: defining class...");
        }
        getClass(classData.getKey(), classData.getValue());
      }
    }
    if (LOG.isLoggable(Level.FINE)) {
      String availableClasses = "no class";
      if (classDatas != null) {
        availableClasses = "";
        for (String className : classDatas.keySet()) {
          availableClasses += className + ", ";
        }
      }
      LOG.fine("BonitaClassLoader: " + this + " created. Datas are available for classDatas : " + availableClasses);
    }
  }

  private Class< ? > getClass(String className, byte[] classData) {
    if (loadedClasses != null && loadedClasses.containsKey(className)) {
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("class " + className + " found in loaded classes.");
      }
      return loadedClasses.get(className);
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Defining class:" + className);
    }
    Class< ? > clazz = defineClass(className, classData, 0, classData.length);
    if (loadedClasses == null) {
      loadedClasses = new HashMap<String, Class< ? >>();
    }
    loadedClasses.put(clazz.getName(), clazz);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Defined class:" + clazz);
    }
    return clazz;
  }

  public Class< ? > findClass(String name) {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Looking for class " + name + "... ");
    }
    try {
      if (classDatas != null && classDatas.containsKey(name)) {
        return getClass(name, classDatas.get(name));
      }
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("class " + name + " not found.");
      }
      return null;
    } catch (Throwable e) {
      throw new BonitaRuntimeException("Problem while loading class with name : " + name, e);
    }
  }

  protected synchronized Class< ? > loadClass(String name, boolean resolve) throws ClassNotFoundException {
    // First check if the class is already loaded
    Class< ? > c = findClass(name);
    if (c == null) {
      c = getParent().loadClass(name);
    }
    if (resolve) {
      resolveClass(c);
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("loadClass: " + name + ", result: " + c);
    }
    return c;
  }

  public Set<String> getClasses() {
    return loadedClasses.keySet();
  }

}


