/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.runtime;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.internal.util.ReflectUtil;
import org.ow2.bonita.definition.ClassInfo;
import org.ow2.bonita.definition.GlobalClassData;
import org.ow2.bonita.definition.Hook;
import org.ow2.bonita.definition.MultiInstantiator;
import org.ow2.bonita.definition.PackageClassData;
import org.ow2.bonita.definition.PerformerAssign;
import org.ow2.bonita.definition.RoleMapper;
import org.ow2.bonita.definition.TxHook;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EnvTool;


public final class ClassDataLoader {

  private static final Logger LOG = Logger.getLogger(ClassDataLoader.class.getName());

  //key = packageDefinitionUUID, value = classLoader
  private static Map<PackageDefinitionUUID, BonitaClassLoader> packageClassLoaders =
    new HashMap<PackageDefinitionUUID, BonitaClassLoader>();

  private static Map<String, BonitaClassLoader> globalClassLoaders = new HashMap<String, BonitaClassLoader>();

  private ClassDataLoader() { }

  public static Class< ? > getClass(PackageDefinitionUUID packageUUID, String className) throws ClassNotFoundException {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Looking for class " + className + ", in package : " + packageUUID);
    }
    Repository repository = EnvTool.getRepository();

    Class< ? > result = lookIntoPackageClassLoader(packageUUID, repository, className);
    if (result != null) {
      return result;
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Class " + className + " not found in packageClassLoaders...");
    }
    result = lookIntoGlobalClassLoader(repository, className);
    if (result != null) {
      return result;
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Class " + className + " not found in globalClassLoaders...");
    }
    //maybe it is present in the current classLoader ? It may be a class delivered in bonita jar ?
    return load(ClassDataLoader.class.getClassLoader(), className);
  }

  protected static Object getInstance(PackageDefinitionUUID packageUUID, ClassInfo classInfo) {
    String className = classInfo.getClassName();
    try {
      Class< ? > clazz = getClass(packageUUID, className);
      return getClassInstance(clazz, classInfo);
    } catch (ClassNotFoundException e) {
      throw new BonitaRuntimeException("The following class was not found : " + className);
    }
  }

  private static Object getClassInstance(Class< ? > clazz, ClassInfo classInfo) {
    Object obj = ReflectUtil.newInstance(clazz);
    if (obj == null) {
      throw new BonitaRuntimeException("Problem while creating a new instance of class : " + clazz.getName()
          + ", returned object is null");
    }
    Map<String, String> parameters = classInfo.getParameters();
    if (parameters != null) {
      for (Map.Entry<String, String> entry : parameters.entrySet()) {
        String fieldName = entry.getKey();
        String fieldvalue = entry.getValue();
        Field field = ReflectUtil.getField(clazz, fieldName);
        ReflectUtil.set(field, obj, fieldvalue);
      }
    }
    return obj;
  }

  private static Class< ? > lookIntoPackageClassLoader(PackageDefinitionUUID packageUUID,
      Repository repository, String className) {
    PackageClassData packageClassData = repository.getPackageClassData(packageUUID);
    if (packageClassData != null) {
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("Class " + className + " is defined in package : " + packageUUID);
      }
      if (!packageClassLoaders.containsKey(packageUUID)) {
        BonitaClassLoader bonitaClassLoader = new BonitaClassLoader(packageClassData.getClasses());
        packageClassLoaders.put(packageUUID, bonitaClassLoader);
      }
      ClassLoader classLoader = packageClassLoaders.get(packageUUID);
      try {
        Class< ? > result = load(classLoader, className);
        if (result != null && result.getClassLoader().equals(classLoader)) {
          return result;
        }
      } catch (ClassNotFoundException e) {
        return null;
      }
    }
    return null;
  }

  public static ClassLoader getGlobalClassLoader(Repository repository, String className) {
    GlobalClassData globalClassData = repository.getGlobalClassData(className);
    if (globalClassData != null) {
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("Class " + className + " is defined as a global class");
      }
      if (!globalClassLoaders.containsKey(className)) {
        Map<String, byte[]> datas = new HashMap<String, byte[]>();
        datas.put(globalClassData.getClassName(), globalClassData.getClassData());
        BonitaClassLoader bonitaClassLoader = new BonitaClassLoader(datas);
        globalClassLoaders.put(className, bonitaClassLoader);
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("BonitaClassLoader added to globalClassLoaders for class " + className);
        }
      }
      return globalClassLoaders.get(className);
    }
    return null;
  }
  
  private static Class< ? > lookIntoGlobalClassLoader(Repository repository, String className) {
    ClassLoader classLoader = getGlobalClassLoader(repository, className);
    if (classLoader != null) {
      try {
        Class< ? > result = load(classLoader, className);
        if (result != null && result.getClassLoader().equals(classLoader)) {
          return result;
        }
      } catch (ClassNotFoundException e) {
        return null;
      }
    }
    return null;
  }

  private static Class< ? > load(ClassLoader classLoader, String className) throws ClassNotFoundException {
    Class< ? > clazz = classLoader.loadClass(className);
    if (classLoader instanceof BonitaClassLoader
        && !BonitaClassLoader.class.isAssignableFrom(clazz.getClassLoader().getClass())) {
      //throw new BonitaRuntimeException("Problem while creating a new instance of class : " + className + ", wrong classLoader");
      return null;
    }
    return clazz;
  }

  public static Object getHookInstance(PackageDefinitionUUID packageUUID, ClassInfo classInfo) {
    Object obj = getInstance(packageUUID, classInfo);
    if (!(obj instanceof Hook) && !(obj instanceof TxHook)) {
      throw new BonitaRuntimeException("The given class : " + classInfo + " is not a Hook instance neither a TxHook instance");
    }
    return obj;
  }
  
  public static RoleMapper getRoleMapperInstance(PackageDefinitionUUID packageUUID, ClassInfo classInfo) {
    Object obj = getInstance(packageUUID, classInfo);
    if (!(obj instanceof RoleMapper)) {
      throw new BonitaRuntimeException("The given class : " + classInfo + " is not a RoleMapper instance");
    }
    return (RoleMapper)obj;
  }

  public static PerformerAssign getPerformerAssignInstance(PackageDefinitionUUID packageUUID, ClassInfo classInfo) {
    Object obj = getInstance(packageUUID, classInfo);
    if (!(obj instanceof PerformerAssign)) {
      throw new BonitaRuntimeException("The given class : " + classInfo + " is not a PerformerAssign instance !");
    }
    return (PerformerAssign)obj;
  }

  public static MultiInstantiator getActivityInstantiatorInstance(PackageDefinitionUUID packageUUID, ClassInfo classInfo) {
    Object obj = getInstance(packageUUID, classInfo);
    if (!(obj instanceof MultiInstantiator)) {
      throw new BonitaRuntimeException("The given class : " + classInfo + " is not a MultiInstantiator instance");
    }
    return (MultiInstantiator)obj;
  }
  
  public static void removePackageClassLoader(PackageDefinitionUUID packageUUID) {
    packageClassLoaders.remove(packageUUID);
  }
  public static void removeGlobalClassLoader(String className) {
    globalClassLoaders.remove(className);
  }
  
  public static Set<String> getPackageClassLoadersClasses() {
    if (packageClassLoaders == null || packageClassLoaders.isEmpty()) {
      return null;
    }
    Set<String> result = new HashSet<String>();
    for (BonitaClassLoader bcl : packageClassLoaders.values()) {
      if (bcl.getClasses() != null) {
        result.addAll(bcl.getClasses());
      }
    }
    return result;
  }
  public static Set<String> getGlobalClassLoadersClasses() {
    if (globalClassLoaders == null || globalClassLoaders.isEmpty()) {
      return null;
    }
    return globalClassLoaders.keySet();
  }
  
  public static void clear() {
    packageClassLoaders.clear();
    globalClassLoaders.clear();
  }
}


