/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.ow2.bonita.env.EnvironmentNotFoundException;
import org.ow2.bonita.facade.APIAccessor;
import org.ow2.bonita.facade.Context;
import org.ow2.bonita.facade.QueryAPIAccessor;
import org.ow2.bonita.facade.ejb.ejb2.EJB2ClientAPIAccessorImpl;
import org.ow2.bonita.facade.ejb.ejb2.EJB2QueryAPIAccessorImpl;
import org.ow2.bonita.facade.ejb.ejb3.EJB3APIAccessorImpl;
import org.ow2.bonita.facade.ejb.ejb3.EJB3QueryAPIAccessorImpl;
import org.ow2.bonita.facade.impl.QueryAPIAccessorImpl;
import org.ow2.bonita.facade.impl.StandardAPIAccessorImpl;
import org.ow2.novabpm.util.Misc;

/**
 *This utility class has been provided to unify access to Bonita APIs and to
 * avoid the use of lookups in JEE deployments:
 * org.ow2.bonita.util.AccessorUtil.<br>
 *Through this class, Nova Bonita APIs can be reached in a unified way in both
 * local and remote applications.<br>
 *For that to be done, the system property called "org.ow2.bonita.api-type"
 * must be defined at client side to specify whether the APIs will be reached
 * locally or remotely<br>
 *(possible values are "standard", "auto-detect", "ejb2" and "ejb3").<br>
 */

public final class AccessorUtil {

  public static final String API_TYPE_PROPERTY = "org.ow2.bonita.api-type";
  public static final String RUNTIMEAPI_JNDINAME = "runtimeAPI";
  public static final String MANAGEMENT_JNDINAME = "managementAPI";
  public static final String DEFINITIONAPI_JNDINAME = "definitionAPI";
  public static final String QUERYRUNTIME_JNDINAME = "queryRuntimeAPI";
  public static final String QUERYDEFINITION_JNDINAME = "queryDefinitionAPI";
  public static final String COMMANDAPI_JNDINAME = "commandAPI";

  private static final Logger LOG = Logger.getLogger(AccessorUtil.class.getName());
  private static final ThreadLocal<Context> CONTEXT = new ThreadLocal<Context>() {
    // Static initialization
    @Override
    protected Context initialValue() {
      final Context context = AccessorUtil.findContext();
      if (AccessorUtil.LOG.isLoggable(Level.INFO)) {
        AccessorUtil.LOG.info("Using: " + context + " as the api-type.");
      }
      return context;
    }
  };

  static final APIAccessor STANDALONE_ACCESSOR = new StandardAPIAccessorImpl();
  static final QueryAPIAccessor STANDALONE_QUERY_ACCESSOR = new QueryAPIAccessorImpl();

  private static InitialContext initialContext;
  static {
    // First initialize the context, next check how to access api
    try {
      AccessorUtil.initialContext = new InitialContext();
    } catch (final NamingException ne) {
      throw new ExceptionInInitializerError(ne);
    }
  }

  static Context findContext() {
    String apiType = System.getProperty(AccessorUtil.API_TYPE_PROPERTY);
    if (apiType == null) {
      if (AccessorUtil.LOG.isLoggable(Level.INFO)) {
        AccessorUtil.LOG.info("Property: " + AccessorUtil.API_TYPE_PROPERTY + " has not been specified for api-type."
            + " Possible values are: "
            + Arrays.asList(Context.values()) + ". Using default value: " + Context.AutoDetect);
      }
      apiType = Context.AutoDetect.toString();
    }
    final Context result = Misc.stringToEnum(Context.class, apiType);
    if (!result.equals(Context.AutoDetect)) {
      if (AccessorUtil.LOG.isLoggable(Level.INFO)) {
        AccessorUtil.LOG.info("API-Type: " + result + " has been specified through property: " + AccessorUtil.API_TYPE_PROPERTY);
      }
      return result;
    }
    return AccessorUtil.autoDetectContext();
  }

  /*
   * Search in which mode we should access to API: First search in the
   * environment. If not present, try to lookup EJB3 RuntimeAPI interface and
   * after EJB2 RuntimeAPI interface if EJB3 interface is not available. Assume
   * that we are in an J2SE mode if none of these are available.
   */
  private static Context autoDetectContext() {
    if (AccessorUtil.LOG.isLoggable(Level.INFO)) {
      AccessorUtil.LOG.info("Auto detecting api-type.");
    }
    try {
      if (AccessorUtil.LOG.isLoggable(Level.INFO)) {
        AccessorUtil.LOG.info("Looking into environment the specified api-type.");
      }
      final Context result = Context.getFromEnvironment();
      if (AccessorUtil.LOG.isLoggable(Level.INFO)) {
        AccessorUtil.LOG.info("API-Type: " + result + " has been taken from the environment."
            + " (Note that the api-type can be set to: " + Context.AutoDetect
            + " in the environment. So this api-type may be the output of the autodetecting algorithm)." 
            + " Increase " + Context.class.getName()
            + " logging level to " + Level.FINER + " for details.");
      }
      return result;
    } catch (final EnvironmentNotFoundException iee) {
      return Context.autoDetect();
    }
  }

  private AccessorUtil() {
  }

  public static Context getContext() {
    return AccessorUtil.CONTEXT.get();
  }

  // The Context of the AccessorUtil should be set to Standard 
  // when inside the engine core!
  // This ensures that hooks (and other JavaCode) get executed 
  // properly when they call AccessorUtil.get*Accessor();
  // Note that the Context in the APIInterceptor remains unchanged 
  // (hence we have two contexts). The one in the APIInterceptor is used
  // to fetch the caller id.
  public static void setContext(final Context context) {
    AccessorUtil.CONTEXT.set(context);
  }

  public static APIAccessor getAPIAccessor() {
    switch (AccessorUtil.getContext()) {
      case EJB2:
        return new EJB2ClientAPIAccessorImpl();
      case EJB3:
        return new EJB3APIAccessorImpl();
      default:
        return AccessorUtil.STANDALONE_ACCESSOR;
    }
  }

  public static QueryAPIAccessor getQueryAPIAccessor() {
    switch (AccessorUtil.getContext()) {
      case EJB2:
        return new EJB2QueryAPIAccessorImpl();
      case EJB3:
        return new EJB3QueryAPIAccessorImpl();
      default:
        return AccessorUtil.STANDALONE_QUERY_ACCESSOR;
    }
  }

  @SuppressWarnings("unchecked")
  public static <T> T lookup(final String name) throws NamingException {
    return (T) AccessorUtil.initialContext.lookup(name);
  }

}
